#include <Arduino.h>
#include <SPI.h>
#include <SD.h>

#include "stopwatch.h"

#include "Experiment.h"

#include "tasks.h"
#include "functions.h"
#include "sched_time.h"

#include "schedule.h"

#include "panic.h"
#include "io.h"

// run each task set for 30 seconds
#define EXP_DURATION 30000000

static unsigned long exp_start_time;

bool experiment_has_finished(void)
{
	unsigned long finish_time = exp_start_time + EXP_DURATION;
	unsigned long now = micros();
	return time_after(now, finish_time);
}

extern StopWatch overhead;

static void write_results(unsigned int folder_id, unsigned int ts_id,
	String &name)
{
	digitalWrite(LED_BUILTIN, HIGH);

	Serial.print(F("Overheads    "));
	Serial.print(F(SCHEDULER_KEY));
	Serial.print(F(", "));
	Serial.print(folder_id);
	Serial.print(F(", "));
	Serial.print(ts_id);
	Serial.print(F(", "));
	Serial.print(overhead.count);
	Serial.print(F(", "));
	Serial.print(overhead.total);
	Serial.print(F(", "));
	Serial.print(overhead.max_val);
	Serial.print(F(", "));
	Serial.println(overhead.min_val);

	Serial.println(name);
	File data_file = SD.open(name, FILE_WRITE);
	BUG_ON(!data_file);

	data_file.print(F(SCHEDULER_KEY));
	data_file.print(F(", "));
	data_file.print(folder_id);
	data_file.print(F(", "));
	data_file.print(ts_id);
	data_file.print(F(", "));
	data_file.print(overhead.count);
	data_file.print(F(", "));
	data_file.print(overhead.total);
	data_file.print(F(", "));
	data_file.print(overhead.max_val);
	data_file.print(F(", "));
	data_file.println(overhead.min_val);
	data_file.close();
	overhead.reset();
}

#ifdef TABLES_IN_FLASH

void start_next_experiment(void)
{
	schedule_init();

	Serial.print(F("Starting experiment: "));
	Serial.println(F(EXP_NAME));

	Serial.print(F("ready to go: "));
	Serial.println(F(SCHEDULER_NAME));
	digitalWrite(LED_BUILTIN, LOW);

	exp_start_time = micros();
	reset_sched_time();
}

void end_experiment(void)
{
	String name = EXP_DATA_FILE;
	write_results(number_of_tasks, EXP_ID, name);
}

#else /* !TABLES_IN_FLASH */

#define TOTAL_EXPERIMENTS 1000
#define TOTAL_FOLDERS 4

#define MAX_IDLE_TIMES       600
#define MAX_IRREGULARITIES   200
#define MAX_TABLE_LENGTH    1300


static unsigned long exp_duration;

static void load_taskset(String fname)
{
	File file = SD.open(fname);

	BUG_ON(!file);

	/* first two lines do not contain tasks */
	skip_lines(file, 1);

	/* second line contains the experiment duration */
	bool exp_ok = parse_value(file, exp_duration, '\n');
	BUG_ON(!exp_ok);
	BUG_ON(!exp_duration);

	/* remaining lines contain tasks */
	while (file.peek() != -1) {
		/* data remains, parse a task */
		unsigned long wcet, deadline, period;
		bool parsed_task_ok = parse_3tuple(file, wcet, deadline, period);
		BUG_ON(!parsed_task_ok);
		BUG_ON(!wcet);
		BUG_ON(wcet > deadline);
		BUG_ON(wcet > period);
		add_task(get_spin_job_for_wcet(wcet), wcet, deadline, period);
	}

	/* make sure we actually loaded something */
	BUG_ON(!number_of_tasks);

	file.close();

	Serial.println(fname);
}

#ifdef WANT_SCHED_OE


static struct idle_time_record idle_times[MAX_IDLE_TIMES];
static struct irregularity_record irregularities[MAX_IRREGULARITIES];

static void load_idle_times(String fname)
{
	BUILD_BUG_ON(sizeof(struct idle_time_record) != 6);

	File file = SD.open(fname);
	struct idle_time_record *recs = idle_times;
	unsigned int i;


	/* parse all the idle times, one per line */
	for (i = 0; i < MAX_IDLE_TIMES && file.peek() != -1; i++) {
		unsigned long start, duration;
		bool parsed_idle_time_ok = parse_2tuple(file, start, duration);
		BUG_ON(!parsed_idle_time_ok);
		recs[i].start = start;
		recs[i].duration = duration;
		/* check for truncation */
		BUG_ON(recs[i].start != start);
		BUG_ON(recs[i].duration != duration);
		/* check for proper ordering */
		BUG_ON(i > 0 && recs[i - 1].end() > recs[i].start);
	}


	/* should have consumed all data */
	bool loaded_all = file.peek() == -1;
	file.close();

	if (!loaded_all) {
		note(fname);
		note(F(" idle-time table overflow\n"));
		Serial.println(F("overflow"));
	}

	oe_set_idle_times(recs, i);

	Serial.println(fname);
}

static void load_irregularities(String fname)
{
	BUILD_BUG_ON(sizeof(struct irregularity_record) != 6);

	unsigned long offset[MAX_TASKS];
	char current_task = -1;

	File file = SD.open(fname);
	struct irregularity_record *recs = irregularities;
	unsigned int i;

	/* parse all the idle times, one per line */
	for (i = 0; i < MAX_IRREGULARITIES && file.peek() != -1; i++) {
		unsigned long task_id, job_id, duration;
		bool parsed_irreg_ok = parse_3tuple(file, task_id, job_id, duration);
		BUG_ON(!parsed_irreg_ok);

		recs[i].job_id = job_id;
		recs[i].delay = duration;
		/* check for truncation */
		BUG_ON(recs[i].job_id != job_id);
		BUG_ON(recs[i].delay != duration);
		/* check for proper ordering */
		BUG_ON(i > 0 && task_id == (byte) current_task &&
			   recs[i - 1].job_id >= recs[i].job_id);
		while (current_task < (char) task_id) {
			current_task++;
			offset[(byte) current_task] = i;
		}
	}

	/* should have consumed all data */
	bool loaded_all = file.peek() == -1;
	file.close();

	if (!loaded_all) {
		note(fname);
		note(F(" irregularity table overflow\n"));
		Serial.println(F("overflow"));
	}

	/* fill in any remaining offsets */
	while (current_task < MAX_TASKS) {
		current_task++;
		offset[(byte) current_task] = i;
	}

	for (byte j = 0; j < number_of_tasks; j++) {
		oe_set_irregularities_for(j, recs + offset[j],
			j + 1 < number_of_tasks ?
			offset[j + 1] - offset[i] :
			i - offset[i]);
	}

	Serial.println(fname);
}
static void load_hyperperiod(String fname)
{
	File file = SD.open(fname);

	unsigned long hp;
	bool parsed_hp_ok = parse_value(file, hp, '\n');

	BUG_ON(!parsed_hp_ok);
	BUG_ON(!hp);

	oe_set_hyperperiod(hp);
	file.close();
	Serial.println(fname);
}

#endif

#ifdef WANT_SCHED_TABLE

static struct table_entry table[MAX_TABLE_LENGTH];

static void load_sched_table(String fname)
{
	/* we don't want to spend ore than 4 bytes per entry */
	BUILD_BUG_ON(sizeof(struct table_entry) != 4);

	File file = SD.open(fname);
	struct table_entry *tab = table;
	unsigned int i;

	/* parse all table entries, one per line */
	for (i = 0; i < MAX_TABLE_LENGTH && file.peek() != -1; i++) {
		unsigned long id, length;
		bool parsed_table_ok = parse_2tuple(file, id, length);
		BUG_ON(!parsed_table_ok);
		BUG_ON(id != 0xff && id >= number_of_tasks);
		tab[i].task_id = id;
		tab[i].slot_length = length;
		/* check for incorrect truncation */
		BUG_ON(tab[i].slot_length != length);
	}

	/* should have consumed all data */
	bool loaded_all = file.peek() == -1;
	file.close();

	if (!loaded_all) {
		note(fname);
		note(F(" scheduling table overflow\n"));
		Serial.println(F("overflow"));
	}

	tt_set_table(tab, i);

	Serial.println(fname);
}

#endif

// this function comes from this address:
// https://learn.adafruit.com/memories-of-an-arduino/measuring-free-memory
// http://playground.arduino.cc/Code/AvailableMemory
int free_ram()
{
	extern int __heap_start, *__brkval;
	int v;
	return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}


static void exp_start(String basename)
{
	clear_tasks();
	schedule_init();

	Serial.print(F("Loading experiment: "));
	Serial.println(basename);
	Serial.print(F("RAM: "));
	Serial.println(free_ram(), DEC);

	load_taskset(basename + "ts.txt");
// 	cat(basename + ".txt");
// 	cat(basename + "hy.txt");
// 	cat(basename + "tt.txt");
// 	cat(basename + "it.txt");
// 	cat(basename + "pi.txt");

#ifdef WANT_SCHED_TABLE
	load_sched_table(basename + "tt.txt");
#endif

#ifdef WANT_SCHED_OE
	load_hyperperiod(basename + "hy.txt");
	load_idle_times(basename + "it.txt");
	load_irregularities(basename + "pi.txt");
#endif

	Serial.print(F("ready to go: "));
	Serial.println(F(SCHEDULER_NAME));
	digitalWrite(LED_BUILTIN, LOW);

	exp_start_time = micros();
	reset_sched_time();
}

static byte folder = 0;
static unsigned int exp_index = TOTAL_EXPERIMENTS;

static String exp_name(void)
{
	String basename = "datan";
	basename += (folder * 3);
	basename += '/';
	basename += exp_index;
	basename += '/';
	return basename;
}

static bool results_file_exists(void)
{
	String name = "RESULTS/DATAN";
	name += (folder * 3);
	name += '/';
	name += exp_index;

	if (!SD.exists(name)) {
		Serial.print("Creating folder'");
		Serial.print(name);
		Serial.println("'");
		SD.mkdir(name);
	}

	name += '/';
	name += SCHEDULER_KEY ".csv";

	return SD.exists(name);
}

void start_next_experiment(void)
{
	String fname;

	do {
		exp_index++;
		if (exp_index > TOTAL_EXPERIMENTS) {
			exp_index = 1;
			folder++;
		}
		/* finish when we've hit all folders */
		BUG_ON(folder > TOTAL_FOLDERS);

	} while (results_file_exists());

	exp_start(exp_name());
}

void end_experiment(void)
{
	String name = "RESULTS/DATAN";
	name += (folder * 3);
	name += '/';
	name += exp_index;
	name += "/" SCHEDULER_KEY ".csv";

	write_results(folder * 3, exp_index, name);
}

#endif  /* !TABLES_IN_FLASH */

