/* An implementation of a non-preemptive table-driven scheduler. */

#include "Arduino.h"

#include "config.h"

#include "panic.h"

#include "tasks.h"
#include "sched_time.h"

#include "sched_table.h"

static unsigned long next_time_to_switch_slots;

#ifdef TABLES_IN_FLASH

/* include statically generated table */
#include "gentab_tt.h"

/* accessor macros --- load from PGM */
#define get_table_slot(i) \
	({			\
		union {		\
			uint32_t raw; \
			struct table_entry slot; \
		} __tmp;	\
		__tmp.raw = pgm_read_dword_near(the_table + i);	\
		__tmp.slot; \
	})

#else /* !TABLES_IN_FLASH */

/* tables loaded into RAM --- just regular pointers */
static table_entry* the_table;
static unsigned int table_length;
/* accessor macro --- direct memory accesses */
#define get_table_slot(i)   the_table[i]

#endif

static unsigned int table_idx;

void tt_init(void)
{
	next_time_to_switch_slots = 0;
	table_idx = 0;
#ifndef TABLES_IN_FLASH
	the_table = NULL;
	table_length = 0;
#endif
}

#ifndef TABLES_IN_FLASH
void tt_set_table(struct table_entry *table, unsigned int length)
{
	the_table = table;
	table_length = length;
}
#endif

sched_decision_t tt_schedule(void)
{
	unsigned long now = current_sched_time();

	/* if no table has been installed, we can't work */
	BUG_ON(!table_length);

	/* check if we need to advance to the next slot */
	if (time_after(now, next_time_to_switch_slots)) {
		/* look at current slot */
		struct table_entry slot = get_table_slot(table_idx);
		/* figure out when to schedule again next */
		next_time_to_switch_slots += slot.slot_length;
		/* figure out which task to schedule now */
		byte i = slot.task_id;
		/* advance in table */
		table_idx++;
		/* check for table wrap-around */
		if (table_idx == table_length)
			table_idx = 0;
		/* finally, return current job if slot is not idle */
		if (i < number_of_tasks)
			return (sched_decision_t) {tasks[i].job_function, 0};
	}

	/* If we reach here, either we either have an idle slot, or
	 * the job in the current slot already completed and we are waiting
	 * for the start of the next slot. Just idle. */
	return NO_JOB_READY;
}
